{###############################
 #                             #
 # Practica 3: Sudoku      G49 #
 #                             #
 # Alejandro Dobeson Beirao    #
 #                             #
 ###############################}

program juegosudoku;

  {Con este programa, podremos jugar al sudoku siempre que tengamos un
   fichero de texto con el tablero original, y la solucion}

uses crt;

const
  MaxFilas = 9;
  MaxColumnas = 9;

  {Mediante estas dos constantes podremos controlar en todo momento como
   de grande sera la tabla que usemos para el juego, siendo esta facilmente
   modificable}

type
  TablaSudoku = array[1..MaxFilas, 1..MaxColumnas] of integer;

  {Aqui declaramos las dimensiones del Sudoku}

  numero = set of 1..9;

  {Declaramos un tipo que sea un conjunto que pueda contener valores
   del 1 al 9, y que utilizaremos mas tarde para las comprobaciones
   de los errores en los sudoku y en los valores que introduza el usuario}

var
  TablaSudokuOriginal,
  TablaSudokuModificado,
  TablaSudokuSolucion: TablaSudoku;

  {Declaramos por separado cada uno de los arrays para el sudoku que usaremos.
   Podria haber sido mas sencillo utilizar solo uno de 3 dimensiones e ir
   variando el valor de la tercera coordenada, pero declarando tres sudokus
   distintos (El introducido originalmente, el que va modificando el usuario
   y el que ya esta completo) se hace bastante mas legible el codigo y es mas
   sencillo comprobar por donde vamos y que estamos haciendo en cada paso}

  Posibilidad: numero;

  {Para un procedimiento, necesitaremos sacar un conjunto con valores de 1 a 9
   con los que denotaremos los posibles valores que puede tomar una casilla.
   Esto servira tanto para el punto 1 como para el punto 6 de la practica}

  Opcion, Fila, Columna: integer;

  {Opcion recogera la eleccion del usuario en el menu
   Fila recogera la fila de la coordenada que quiere modificar el usuario
   Columna hara lo mismo que la anterior, pero con lo propio}

{############################################################################}

procedure menu (var aux: integer);

  {Este procedimiento colocara y mostrara por ventana el menu, recogiendo
   y devolviendo despues la opcion elegida por el usuario}

begin
  window(1, 16, 80, 43);
  writeln;
  writeln('1. Ver los posibles valores de una casilla');
  writeln('2. Colocar un valor en una casilla');
  writeln('3. Borrar valor de una casilla');
  writeln('4. Detectar valores incorrectos');
  writeln('5. Reiniciar el tablero');
  writeln('6. Rellenar celdas simples');
  writeln;
  writeln('0. Salir');

  repeat
    writeln;
    write('Haz tu eleccion: ');
    read(aux);
    if (aux < 0) or (aux > 6) then
      writeln('Introduzca una opcion valida');
  until (aux >= 0) and (aux <= 6)

  {Con esta sencilla comprobacion en forma de repeat, verificamos que el
   usuario meta una opcion valida, y que no cause posibles problemas.}

end;

 (***************************)

procedure crearsudoku (var Sudoku, SudokuMod, SudokuSol: TablaSudoku);

  {Este procedimiento es el codigo suministrado al principio de la practica.
   Lo utilizaremos para rellenar los valores que despues formaran parte del
   array original, array modificado y array solucion del sudoku}

  var

    f, c, v:  integer;
      (* f = fila de la casilla; c = columna de la casilla; v = valor *)
    nombreFichero: string;
      (* cadena para guardar el nombre y la ruta del fichero *)
    fich: text;
      (* fichero logico *)

  begin

    repeat
      write('Introduzca el nombre del tablero: ');

      {Modificacion minima. Usaremos dos veces el mismo codigo. Una para
       pedir el tablero original, y otra para pedir la solucion. Por lo
       que cambiamos tambien la frase para saber que pedimos en cada
       momento.}

      readln(nombreFichero);

      assign(fich, nombreFichero);

      {$I-}
      reset(fich);
      {$I+}

    until IOResult = 0;

    while not eof(fich) do
      begin
        readln(fich, f, c, v);

        {Una vez tenemos leido el valor, solo hace falta localizarlo dentro
         del array original.}

        Sudoku[f, c] := v;
      end;

    {Nuestro array modificado parte de los mismos valores que el original,
     asi que, en vez de leer de nuevo todos los valores, igualamos el array
     al que ya tenemos relleno.}

    SudokuMod := Sudoku;

    close(fich);

    repeat
      write('Introduzca el nombre de la solucion: ');

      {Comienza la segunda parte del procedimiento. El codigo es exactamente
       igual al anterior, pero esta vez pediremos la solucion, y la usaremos
       para rellenar el array solucion.}

      readln(nombreFichero);

      assign(fich, nombreFichero);

      {$I-}
      reset(fich);
      {$I+}

    until IOResult = 0;

    while not eof(fich) do
      begin
        readln(fich, f, c, v);

        {Una vez leido el valor, lo localizamos en el array.}

        SudokuSol[f, c] := v;
      end;

    close(fich);

  clrscr

  end;

 (***************************)

procedure EscribeCasillero;

  {Este procedimiento escribira el casillero en el que iremos escribiendo
   despues el sudoku en si. Debido al diseo, pensado de antemano, solo
   necesitaremos dibujarlo una vez, ya que el resto de veces nos limitaremos
   a tener cuidado de no permitir la escritura encima del dibujo.}

var
  aux: integer;

  {Usaremos aux para llevar una doble cuenta. En primer lugar, para los for
   que dibujan el tablero, y en segundo lugar, para escribir la coordenada
   de la fila correspondiente a cada una.}

begin
  window(31, 1, 46, 14);
  writeln('   123 456 789');
  writeln('  ͻ');
  for aux := 1 to 3 do
    writeln(aux, '          ');
  writeln('  ͹');
  for aux := 4 to 6 do
    writeln(aux,'          ');
  writeln('  ͹');
  for aux := 7 to 9 do
    writeln(aux,'          ');
  write('  ͼ');
end;

 (***************************)

procedure EscribeSudoku(SudokuOriginal, SudokuModificado: TablaSudoku);

  {Este es el procedimiento que reescribira el sudoku a cada vez que se
   modifique. Introduciremos una pequea parte de codigo para diferenciar
   entre los valores originales y los valores modificados por el usuario,
   escribiendo estos ultimos en color blanco, para que se vea de un vistazo
   cuales son las casillas modificables. Tambien meteremos otro comprobador
   para no escribir las casillas con un 0 en el array y dejarlas en blanco,
   para que quede mas estetico.}

var

  f, ff, c, cc: integer;

  {f y c corresponden a las variables fila y columna del array, paralelamente
   a ff y cc, que corresponden a la fila y columna de posicion en pantalla.}

begin

  ff := 1;
  cc := 1;
  window(34, 3, 45, 13);

  {Lo primero que haremos sera colocarnos en la ventana que nos viene mejor
   para redibujar el sudoku, y mediante dos bucles for, haremos un recorrido
   completo por toda la pantalla y el array modificado, aadiendo los valores
   originales en color azul, los metidos por el usuario en verde y los
   espacios vacios (ceros) con un espacio en blanco.}

  for f := 1 to MaxFilas do
    begin
      for c := 1 to MaxColumnas do
        begin
          gotoxy(cc, ff);
          if SudokuOriginal[f, c] <> 0 then
            begin
              textcolor(blue);
              write(SudokuOriginal[f, c]);
              normvideo;
            end;
          if SudokuModificado[f, c] = 0 then
            write(' ');
          if (SudokuOriginal[f, c] = 0) and (SudokuModificado[f, c] <> 0) then
            begin
              textcolor(green);
              write(SudokuModificado[f, c]);
              normvideo;
            end;
          cc := cc + 1;
          if (cc = 4) or (cc = 8) then
            cc := cc + 1;

            {Este comprobador es bastante importante. Lo que hace es que,
             siempre que lleguemos a una coordenada en la que hay dibujada
             una linea del casillero, hacemos que la coordenada siguiente a
             escribir sea la siguiente, y no esa. Haremos lo mismo con las
             filas.}

        end;
      cc := 1;
      ff := ff + 1;
      if (ff = 4) or (ff = 8) then
        ff := ff + 1;
    end

end;

 (***************************)

procedure Comprobador(SudokuMod: TablaSudoku;
                      f, c: integer;
                      var x, y, z: numero);

  {Este procedimiento utiliza el array modificado y las coordenadas de la
   casilla a estudiar, y devuelve tres conjuntos, x, y, z, que nos serviran
   para determinar los posibles valores que podra tener la casilla.
   Respectivamente, x comprobara los valores de la fila, y los valores de la
   columna, y z los valores de la cuadricula.}

var
  aux, ff, cc: integer;

  {Estas tres variables nos serviran para hacer los recorridos necesarios para
   cada rellenar las tres pruebas, y limitar el recorrido del array modificado
   a la cuadricula que nos interesa.}

begin
  x := [];
  y := [];
  z := [];

  {Inicializamos los tres conjuntos como conjuntos vacios para evitar fallos.}

  for aux := 1 to 9 do
    x := x + [SudokuMod[f, aux]];

  {Aqui obtenemos los valores de x. Dejamos la fila fija, la cual obtuvimos
   fuera de este procedimiento, y hacemos un recorrido por todos los valores
   posibles de columnas}

  for aux := 1 to 9 do
    y := y + [SudokuMod[aux, c]];

  {Aqui obtenemos los valores de y. Dejamos la columna fija, la cual obtuvimos
   fuera de este procedimiento, y hacemos un recorrido por todos los valores
   posibles de filas}

  case f of
    1, 2, 3: case c of
               1, 2, 3: begin
                          for cc := 1 to 3 do
                            for ff := 1 to 3 do
                              z := z + [SudokuMod[ff, cc]];
                        end;
               4, 5, 6: begin
                          for cc := 4 to 6 do
                            for ff := 1 to 3 do
                              z := z + [SudokuMod[ff, cc]];
                        end;
               7, 8, 9: begin
                          for cc := 7 to 9 do
                            for ff := 1 to 3 do
                              z := z + [SudokuMod[ff, cc]];
                        end;
               end;
    4, 5, 6: case c of
               1, 2, 3: begin
                          for cc := 1 to 3 do
                            for ff := 4 to 6 do
                              z := z + [SudokuMod[ff, cc]];
                        end;
               4, 5, 6: begin
                          for cc := 4 to 6 do
                            for ff := 4 to 6 do
                              z := z + [SudokuMod[ff, cc]];
                        end;
               7, 8, 9: begin
                          for cc := 7 to 9 do
                            for ff := 4 to 6 do
                              z := z + [SudokuMod[ff, cc]];
                        end;
             end;
    7, 8, 9: case c of
               1, 2, 3: begin
                          for cc := 1 to 3 do
                            for ff := 7 to 9 do
                              z := z + [SudokuMod[ff, cc]];
                        end;
               4, 5, 6: begin
                          for cc := 4 to 6 do
                            for ff := 7 to 9 do
                              z := z + [SudokuMod[ff, cc]];
                        end;
               7, 8, 9: begin
                          for cc := 7 to 9 do
                            for ff := 7 to 9 do
                              z := z + [SudokuMod[ff, cc]];
                        end;
             end;
    end;

    {Mediante dos cases anidados, limitamos el recorrido del array, fijando
     la cuadricula mediante la fila y columna obtenidas fuera.}

end;

 (***************************)

procedure Coordenadas(Sudoku: TablaSudoku; var f, c: integer);

  {Este procedimiento preguntara al usuario y comprobara las coordenadas de
   la casilla a modificar. Hace una comprobacion a priori para ver si es
   posible, y otra a posteriori para comprobar si se puede modificar o viene
   dada por defecto.}

var
  fallo: boolean;

  {Fallo sera una variable de control para el bucle. Sera falsa si la casilla
   es modificable, o verdadera si no se puede modificar, y entonces pedira
   otras coordenadas.}

begin
  writeln;

  repeat
  fallo := false;

    repeat
      write('Escribe fila: ');
      readln(f);
      if (f < 1) or (f > MaxFilas) then
        begin
          writeln('Fila incorrecta. Verifique su opcion.');
          writeln;
        end
    until (f >= 1) and (f <= MaxFilas);

    repeat
      write('Escribe columna: ');
      readln(c);
      if (c < 1) or (c > MaxColumnas) then
        begin
          writeln('Columna incorrecta. Verifique su opcion.');
          writeln;
        end
    until (c >= 1 ) and (c <= MaxColumnas);

    if (Sudoku[f, c] <> 0) then
      begin
        writeln('Esa casilla no se puede editar.');
        writeln;
        fallo := true;
      end

  until fallo = false;
end;

 (***************************)

procedure Posibilidades(SudokuMod: TablaSudoku;
                        f, c: integer;
                        var pos: numero);

  {Este procedimiento simplifica lo obtenido en el procedimiento Comprobador,
   uniendolo en un solo conjunto con todos los valores no posibles, y rellena
   otro, con todos los valores que si podrian escribirse, el cual devuelve.}

var
  i: integer;
  pruebax, pruebay, pruebaz, pruebaj: numero;

  {i es una variable auxiliar, pruebax/y/z es de las mismas caracteristicas
   que x/y/z del procedimiento Comprobador, y se aade la pruebaj. Esta sera
   la suma de las otras tres.}

begin
  pos := [];
  Comprobador(SudokuMod, f, c, pruebax, pruebay, pruebaz);
  pruebaj := pruebax + pruebay + pruebaz;
  for i := 1 to 9 do
    begin
      if ((i in pruebaj) = false) then
        pos := pos + [i]

        {De esta forma, rellenamos pos (futuro Posibilidades) con los valores
         que si pueden ir dentro de esa casilla.}

    end;
end;

 (***************************)

procedure EscribePosibilidades(Pos: numero);

  {Este procedimiento obtiene las posibilidades de la casilla estudiada en
   pasos anteriores, y los muestra por pantalla.}

var
  i: integer;

  {i es una variable auxiliar para hacer el "recorrido" del conjunto.}

begin
  if (Pos <> []) then
    begin
      writeln('Los posibles valores son:');
      writeln;
      for i := 1 to 9 do
        if ((i in pos) = true) then
          writeln(i);
      writeln;
      writeln('Pulsa INTRO para continuar...');
      readln;
    end
end;

 (***************************)

procedure ColocarValor(Sudoku: TablaSudoku; var SudokuMod: TablaSudoku);

  {Este procedimiento, que necesita del array original y del modificado,
   devolviendo este ultimo, hace las comprobaciones necesarias y llama
   a los procedimientos pertinentes para poder aadir un valor a una
   casilla.}

var
  f, c, valor, aux: integer;

  {Como en el resto de procedimientos, f es la fila, c la columna, valor es
   autoexplicativo, y aux es una variable auxiliar.}

  pruebax, pruebay, pruebaz: numero;

  {Como hemos visto en otros procedimientos, estas tres variables contendran
   los valores que hay en la misma fila, columna y cuadricula de la casilla
   seleccionada.}

  fallo: boolean;

  {Variable de control para repetir el procedimiento si el valor no es valido
   o ya esta escrito en la misma fila, columna o casilla.}

begin
  Coordenadas(Sudoku, f, c);
  Comprobador(SudokuMod, f, c, pruebax, pruebay, pruebaz);
  repeat
    fallo := false;
    write('Escribe valor: ');
    readln(valor);
    if (valor < 1) or (valor > 9) then
      begin
        writeln('Valor incorrecto. Verifique su opcion');
        fallo := true;
        writeln;
      end
    else
      begin
        if (valor IN pruebax) = FALSE then
           if (valor IN pruebay) = FALSE then
            if (valor IN pruebaz) = FALSE then
              SudokuMod[f, c] := valor
            else
              begin
                writeln('El valor se repite en la misma cuadricula, no es valido.');
                writeln;
                fallo := true
                end
          else
            begin
              writeln('El valor se repite en la misma columna, no es valido.');
              writeln;
              fallo := true
            end
        else
          begin
            writeln('El valor se repite en la misma fila, no es valido.');
            writeln;
            fallo := true
          end
      end
  until fallo = false;
end;

 (***************************)

procedure BorrarValor(Sudoku: TablaSudoku; var SudokuMod: TablaSudoku);

  {Procedimiento sencillo en el que se comprueba que la casilla se puede
   modificar, y se pone a 0 el valor que contiene. Es decir, se borra.}

var
  f, c: integer;

begin
  Coordenadas(Sudoku, f, c);
  SudokuMod[f, c] := 0;
end;

 (***************************)

procedure Errores(Sudoku, SudokuSol, SudokuMod: TablaSudoku;
                  var op: integer);

  {Procedimiento para comprobar los errores del Sudoku. Para ello,
   necesitamos del original, para no comprobar esas casillas, el modificado,
   para hacer la comprobacion en las que si esten modificadas, y la solucion,
   para tener algo con lo que compararlo. Reescribiremos los fallos en rojo
   y los aciertos en verde, y aparte, notificaremos los fallos por escrito.
   Tambien recoge y devuelve la variable op, correspondiente a la opcion del
   menu. Si el sudoku tiene errores o no esta completo, entrara y saldra con
   el mismo valor. En cambio, si el sudoku esta ya completo, esta variable
   tomara el valor 0, indicando asi que el programa debe finalizar.}

var
  f, c, aux: integer;

  {Variables para fila, columna, y mantener una coherencia al escribir
   los fallos.}

  fallos: boolean;

  {Variable usada para distinguir entre un sudoku sin completar o con fallos,
   y un sudoku completo y sin fallos.}

begin
  writeln;
  aux := 29;
  fallos := false;
  for f := 1 to MaxFilas do
    begin
      for c := 1 to MaxColumnas do
        begin
          if Sudoku[f, c] = 0 then
            begin
              if SudokuMod[f, c] = SudokuSol[f, c] then
                begin
                  window(34, 3, 45, 13);
                  case f of
                    1, 2, 3: case c of
                               1, 2, 3: gotoxy(c, f);
                               4, 5, 6: gotoxy(c + 1, f);
                               7, 8, 9: gotoxy(c + 2, f);
                             end;
                    4, 5, 6: case c of
                               1, 2, 3: gotoxy(c, f + 1);
                               4, 5, 6: gotoxy(c + 1, f + 1);
                               7, 8, 9: gotoxy(c + 2, f + 1);
                             end;
                    7, 8, 9: case c of
                               1, 2, 3: gotoxy(c, f + 2);
                               4, 5, 6: gotoxy(c + 1, f + 2);
                               7, 8, 9: gotoxy(c + 2, f + 2);
                             end;
                  end;
                  textcolor(green);
                  write(SudokuMod[f, c]);
                  window(1, aux, 80, 43);
                  normvideo;
                end;
              if (SudokuMod[f, c] <> SudokuSol[f, c]) then
                begin
                  fallos := true;
                  if (SudokuMod[f, c] <> 0) then
                    begin
                      window(34, 3, 45, 13);
                      case f of
                        1, 2, 3: case c of
                                   1, 2, 3: gotoxy(c, f);
                                   4, 5, 6: gotoxy(c + 1, f);
                                   7, 8, 9: gotoxy(c + 2, f);
                                 end;
                        4, 5, 6: case c of
                                   1, 2, 3: gotoxy(c, f + 1);
                                   4, 5, 6: gotoxy(c + 1, f + 1);
                                   7, 8, 9: gotoxy(c + 2, f + 1);
                                 end;
                        7, 8, 9: case c of
                                   1, 2, 3: gotoxy(c, f + 2);
                                   4, 5, 6: gotoxy(c + 1, f + 2);
                                   7, 8, 9: gotoxy(c + 2, f + 2);
                                end;
                      end;
                      textcolor(red);
                      write(SudokuMod[f, c]);
                      normvideo;
                      window(1, aux, 80, 43);
                      writeln('El valor introducido en la fila ', f, ', columna ', c, ' no es correcto');
                      aux := aux + 1;
                    end;
                end;
            end;
        end;
    end;
    writeln;
    if (fallos = true) then
      begin
        writeln('Todos los valores en verde estan en la posicion correcta :)');
        writeln;
        writeln('Pulsa INTRO para continuar...');
        repeat
        until keypressed;
        window(1, 26, 80, 43);
      end
    else
      begin
        writeln('                   ');
        writeln('                               ');
        writeln('                      ');
        writeln('                              ');
        writeln('                   ');
        writeln;
        writeln('Has completado el sudoku!');
        op := 0;
        repeat
        until keypressed;
      end
end;

 (***************************)

procedure AutoCompletar(var SudokuMod: tablasudoku);

  {Este procedimiento detecta las llamadas "celdas simples", es decir,
   aquellas coordenadas del array que solo admiten un solo valor, y guarda
   temporalmente su valor para despues ponerselo al array modificable. El
   porque de esta separacion es la intencion de evitar que vayan surgiendo
   mas celdas simples de las que habia al principio de la comprobacion, por
   efecto de nuestro paso y consecuente rellenado del array modificable. Por
   eso es necesario un segundo array temporal que guarde estos valores, y
   permita seguir realizando todas las comprobaciones sobre el tablero, tal y
   como estaba al comienzo de la comprobacion.}

var
  f, c, i: integer;

  {F para fila, c para columna, i es una auxiliar y contador, lleva la cuenta
   de cuantas casillas estan rellenas.}

  pos: numero;

  {Conjunto con las posibilidades}

  arrayaux: TablaSudoku;

  {En este array temporal guardaremos los valores de las casillas simples, y
   su posicion, para despues colocarlas todas de una sola vez y evitar la
   aparicion de nuevas "casillas simples" a nuestro paso.}

begin
  arrayaux := SudokuMod;
  for f := 1 to MaxFilas do
    begin
      for c := 1 to MaxColumnas do
        begin
          if SudokuMod[f, c] = 0 then
            begin
              pos := [];
              Posibilidades(arrayaux, f, c, pos);
              for i := 1 to 9 do
                if pos = [i] then
                  SudokuMod[f, c] := i
            end
        end
    end;
end;

{############################################################################}

function SudokuCompleto(SudokuMod: TablaSudoku): Boolean;

  {Esta funcion indica al ordenador si el tablero del Sudoku esta completo, o
   no. Para ello, usa el array modificado por el usuario, y devuelve un valor
   booleano}

var
  i, j, contador: integer;

  {i y j son variables auxiliares usadas en el recorrido por todo el array.
   Contador es la variable en la que guardaremos la cantidad de casillas
   rellenas}

begin
  contador := 0;
  for i := 1 to MaxFilas do
    for j := 1 to MaxColumnas do
      if SudokuMod[i, j] <> 0 then
        contador := contador + 1;
  if contador = (MaxFilas * MaxColumnas) then
    begin
      SudokuCompleto := true;
    end
  else
    SudokuCompleto := false;
end;

{############################################################################}

begin
  textmode(co80 or font8x8);
  clrscr;
  writeln('#####################');
  writeln;
  writeln(' Bienvenido a SUDOKU');
  writeln;
  writeln('#####################');
  writeln;
  CrearSudoku(TablaSudokuOriginal, TablaSudokuSolucion, TablaSudokuModificado);
  EscribeCasillero;
  repeat
    EscribeSudoku(TablaSudokuOriginal, TablaSudokuModificado);
    if (SudokuCompleto(TablaSudokuModificado) = true) then
      begin
        opcion := 0;
      end
    else
      begin
        menu(opcion);
        case opcion of
          1: begin
               coordenadas(TablaSudokuOriginal,
                           Fila,
                           Columna);
               posibilidades(TablaSudokuModificado,
                             Fila,
                             Columna,
                             Posibilidad);
               escribeposibilidades(Posibilidad);
               clrscr;
             end;
          2: begin
               colocarvalor(TablaSudokuOriginal,
                            TablaSudokuModificado);
               clrscr;
             end;
          3: begin
               borrarvalor(TablaSudokuOriginal,
                           TablaSudokuModificado);
               clrscr;
             end;
          4: begin
               errores(TablaSudokuOriginal,
                       TablaSudokuSolucion,
                       TablaSudokuModificado,
                       Opcion);
               clrscr;
             end;
          5: begin
               TablaSudokuModificado := TablaSudokuOriginal;
               clrscr;
             end;
          6: begin
               autocompletar(TablaSudokuModificado);
               clrscr;
             end;
        end
      end
  until (opcion = 0);
end.